#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#ducting - code is complex and might fail after some time if parameters is wrong (time waisting). Sensable values are also checked. Gives printout of all wrong parameters. Dirty images alone has x parametrs.

import numpy as np
import xarray as xr
import dask.array as da
from ._zernike_polynomials import _generate_zernike_surface
from ._math_utils import _compute_rot_coords
import time
import itertools
from numba import jit
  
# Do we need the abs for apeture_parms['zernike_size'].
#    apeture_parms['cell_size'] = 1/(beam_parms['cell_size']*beam_parms['image_size']*apeture_parms['oversampling'])
#    apeture_parms['zernike_size'] = np.floor(np.abs(Dish_Diameter*eta/(apeture_parms['cell_size']*lmbd))) #Why is abs used?
#2.) Currently in ARD-20 the Zernike grid parameters are caculated using (in ZernikeCalc.cc):
#size_zernike = floor((D*eta)/dx)
#delta_zernike = 2.0/size_zernike
#
#By reversing the order the delta of the zernike grid and uv grid can match exactly (no floor operation). The zernike grid will be slightly larger (all cells outside the r=1 will still be nulled).
#delta_zernike = (2.0*dx)/(D*eta)
#size_zernike = ceil((D*eta)/dx)
#Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.

def _calc_ant_jones(list_zpc_dataset,j_freq,j_pa,beam_parms):
    pa_prev = -42.0
    freq_prev = -42.0
    i_model_prev = -42
    c = 299792458
    
    
    j_planes = np.zeros((len(list_zpc_dataset),len(j_pa),len(j_freq),len(beam_parms['needed_pol']),beam_parms['image_size'][0],beam_parms['image_size'][1]),np.complex128)
    j_planes_shape = j_planes.shape
    iter_dims_indx = itertools.product(np.arange(j_planes_shape[0]), np.arange(j_planes_shape[1]),np.arange(j_planes_shape[2]))
    ic = beam_parms['image_size']//2 #image center pixel
    
    
    for i_model, i_pa, i_chan in iter_dims_indx:
        #print(i_model,i_pa,i_chan)
        pa = j_pa[i_pa]
        beam = list_zpc_dataset[i_model]
        freq = j_freq[i_chan]

        if (i_model != i_model_prev) or (freq != freq_prev):
            beam_interp = beam.interp(chan=freq,method=beam_parms['zernike_freq_interp'])

        dish_diam = beam.dish_diam
        lmbd = c/freq
        eta = beam_interp.ETA[0,0].values #Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.
        uv_cell_size = 1/(beam_parms['cell_size']*beam_parms['image_size'])
        zernike_cell = (2.0*uv_cell_size*lmbd)/(dish_diam*eta)
        
        
        if (pa != pa_prev) or (freq != freq_prev) :
            beam_parms['parallactic_angle'] = pa
            image_size = (np.ceil(np.abs(2.0/zernike_cell))).astype(int)
            x_grid, y_grid = _compute_rot_coords(image_size,zernike_cell,pa)
            
            r_grid = np.sqrt(x_grid**2 + y_grid**2)
            
            zernike_size = np.array(x_grid.shape)
        
            ic_z = zernike_size//2
            include_last = (zernike_size%2).astype(int)
        
        #assert zernike_size[0] < beam_parms['conv_size'][0] and zernike_size[1] < gcf_parms['conv_size'][1], "The convolution size " + str(gcf_parms['conv_size']) +" is smaller than the aperture image " + zernike_size + " . Increase conv_size"
        
        start = time.time()
        for i_pol,pol in enumerate(beam_parms['needed_pol']):
            a = _generate_zernike_surface(beam_interp.ZC.data[pol,:].compute(),x_grid,y_grid)
            a[r_grid > 1] = 0
            j_planes[i_model, i_pa, i_chan,i_pol,ic[0]-ic_z[0]:ic[0]+ic_z[0]+include_last[0],ic[1]-ic_z[1]:ic[1]+ic_z[1]+include_last[1]] = a
            j_planes[i_model, i_pa, i_chan, i_pol,:,:] = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(j_planes[i_model, i_pa, i_chan,i_pol,:,:])))/(beam_parms['image_size'][0]*beam_parms['image_size'][1])
        #print('One pol set',time.time()-start)
        
        #Normalize Jones
        if 3 not in beam_parms['needed_pol']:
            P_max = np.abs(j_planes[i_model, i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[4]//2,j_planes_shape[5]//2])
            Q_max = P_max
        elif 0 not in beam_parms['needed_pol']:
            Q_max = np.abs(j_planes[i_model, i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[4]//2,j_planes_shape[5]//2])
            P_max = Q_max
        else:
            P_max = np.abs(j_planes[i_model, i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[4]//2,j_planes_shape[5]//2])
            Q_max = np.abs(j_planes[i_model, i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[4]//2,j_planes_shape[5]//2])

        j_planes[i_model, i_pa, i_chan,:,:,:] = j_planes[i_model, i_pa, i_chan,:,:,:]*2/(P_max+Q_max)
        
        pa_prev = pa
        freq_prev = freq
        i_model_prev = i_model
    return j_planes#np.zeros((1,4,2048,2048),dtype=np.complex128)



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from astropy.utils import iers
from astropy.utils import data
from astropy.time import Time
iers_b = iers.IERS_B.open(data.download_file(iers.IERS_B_URL, cache=True))
iers_a = iers.IERS_A.open(data.download_file(iers.IERS_A_URL, cache=True))
iers_auto = iers.IERS_Auto.open()
import numpy as np

def calc_uvw(tel_xds, time_str, phase_center_ra_dec, uvw_parms, check_parms=True):
    """
    Calculate uvw coordinates.
    
    Parameters
    ----------

    Returns
    -------
    uvw : np.array
    """
    from ._parm_utils._check_uvw_parms import _check_uvw_parms
    from ._sirius_utils._array_utils import _calc_baseline_indx_pair
    import copy
    
    _uvw_parms = copy.deepcopy(uvw_parms)
    if check_parms:
        assert(_check_uvw_parms(_uvw_parms)), "######### ERROR: calc_uvw uvw_parms checking failed."
    
    n_ant = tel_xds.dims['ant_name']
    antenna1,antenna2=_calc_baseline_indx_pair(n_ant,_uvw_parms['auto_corr'])

    if _uvw_parms['calc_method'] == 'astropy':
        uvw = calc_uvw_astropy(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    elif _uvw_parms['calc_method'] == 'casa_thread_unsafe':
        uvw = calc_uvw_casa(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    elif _uvw_parms['calc_method'] == 'casa':
        uvw = calc_uvw_casacore(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    #elif _uvw_parms['calc_method'] == 'calc11':
        #uvw = calc_uvw_casa(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2)
    return uvw, antenna1, antenna2


def calc_uvw_astropy(tel_xds, time_str, phase_center_ra_dec, antenna1, antenna2):
    from astropy.time import Time
    import astropy.coordinates as coord
    import astropy.time
    import astropy.units as u
    
    n_time = len(time_str)
    n_ant = tel_xds.dims['ant_name']

    # Time of observation:
    time_str = np.tile(time_str[:,None],(1,n_ant))
    mjd = astropy.time.Time(Time(time_str, scale='utc'), format='mjd', scale='utc')

    # Format antenna positions and array center as EarthLocation.
    ant_pos = np.tile(tel_xds.ANT_POS.values[None,:,:]*u.m,(n_time,1,1))
    antpos_ap = coord.EarthLocation(x=ant_pos[:,:,0], y=ant_pos[:,:,1], z=ant_pos[:,:,2])
    #tel_site = coord.EarthLocation.of_site(site)
    tel_site = coord.EarthLocation(x=tel_xds.site_pos[0]['m0']['value']*u.m, y=tel_xds.site_pos[0]['m1']['value']*u.m, z=tel_xds.site_pos[0]['m2']['value']*u.m)
    
    # Convert antenna pos terrestrial to celestial.  For astropy use
    # get_gcrs_posvel(t)[0] rather than get_gcrs(t) because if a velocity
    # is attached to the coordinate astropy will not allow us to do additional
    # transformations with it (https://github.com/astropy/astropy/issues/6280)
    tel_site_p, tel_site_v = tel_site.get_gcrs_posvel(mjd)
    antpos_c_ap = coord.GCRS(antpos_ap.get_gcrs_posvel(mjd)[0],
            obstime=mjd, obsgeoloc=tel_site_p, obsgeovel=tel_site_v)

    phase_center_ra_dec = coord.SkyCoord(phase_center_ra_dec[:,0]*u.rad, phase_center_ra_dec[:,1]*u.rad, frame='icrs')

    #frame_uvw = phase_center_ra_dec.skyoffset_frame() # ICRS
    frame_uvw = phase_center_ra_dec.transform_to(antpos_c_ap).skyoffset_frame() # GCRS

    # Rotate antenna positions into UVW frame.
    antpos_uvw_ap = antpos_c_ap.transform_to(frame_uvw).cartesian
    
    ant_uvw = np.array([antpos_uvw_ap.y,antpos_uvw_ap.z,antpos_uvw_ap.x])
    ant_uvw = np.moveaxis(ant_uvw, 0, -1)
    
    uvw = np.ascontiguousarray(ant_uvw[:,antenna1,:] - ant_uvw[:,antenna2,:])
    return uvw
             

def calc_uvw_casa(tel_xds, time_str, phase_center_ra_dec,antenna1, antenna2):
    print("Warning CASA uvw code is not thread safe.")

    try:
        import casatools
    except ImportError:
        print('casatools not installed')
        return

    def casa_to_astropy(c):
        import astropy.coordinates as coord
        import astropy.units as u
        """Convert CASA spherical coords to astropy CartesianRepresentation"""
        sph = coord.SphericalRepresentation(
                lon=c['m0']['value']*u.Unit(c['m0']['unit']),
                lat=c['m1']['value']*u.Unit(c['m1']['unit']),
                distance=c['m2']['value']*u.Unit(c['m2']['unit']))
        return sph.represent_as(coord.CartesianRepresentation)

    n_time = len(time_str)
    n_baseline = len(antenna1)
    
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    
    qa = casatools.quanta()
    qq = qa.quantity
    
    me = casatools.measures()
    
    tel_site = tel_xds.site_pos[0]#me.observatory(tel_xds.telescope_name)
    
    ant_pos = tel_xds.ANT_POS.values
    # Format antenna positions for CASA:
    antpos_casa = me.position('ITRF',
                qq(ant_pos[:,0],'m'),
                qq(ant_pos[:,1],'m'),
                qq(ant_pos[:,2],'m'))
    me.done()
    
    uvw = np.zeros((n_time,n_baseline,3))
    
    for i,t in enumerate(time_str):
        ra_dec = phase_center_ra_dec[i//f_pc_time,:]
        me = casatools.measures()
        me.doframe(tel_site)
        me.doframe(me.epoch('utc', str(t)))
        me.doframe(me.direction('J2000',qq(ra_dec[0], 'rad'),qq(ra_dec[1], 'rad')))
        
        # Converts from ITRF to "J2000":
        antpos_c_casa = me.asbaseline(antpos_casa)
        # Rotate into UVW frame
        antpos_uvw_casa = me.touvw(antpos_c_casa)[0]
        me.done()
        
        ant_uvw = casa_to_astropy(antpos_uvw_casa)
        ant_uvw = ant_uvw.xyz.value.T

        uvw[i,:,:] = np.ascontiguousarray(ant_uvw[antenna1,:] - ant_uvw[antenna2,:])
    return uvw
     
###################
def calc_uvw_casacore(tel_xds, time_str, phase_center_ra_dec,antenna1, antenna2):
    def casa_to_astropy(c):
        import astropy.coordinates as coord
        import astropy.units as u
        """Convert CASA spherical coords to astropy CartesianRepresentation"""
        sph = coord.SphericalRepresentation(
                lon=c['m0']['value']*u.Unit(c['m0']['unit']),
                lat=c['m1']['value']*u.Unit(c['m1']['unit']),
                distance=c['m2']['value']*u.Unit(c['m2']['unit']))
        return sph.represent_as(coord.CartesianRepresentation)

    import pkg_resources
    import os
    n_time = len(time_str)
    n_baseline = len(antenna1)
    
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    
    
    from casacore.measures import measures
    from casacore.quanta import quantity as qq

    me = measures()

    #print(tel_xds)
    tel_site = tel_xds.site_pos[0] #me.observatory(name=tel_xds.telescope_name)
    
    ant_pos = tel_xds.ANT_POS.values
    # Format antenna positions for CASA:
    antpos_casa = me.position('ITRF',
                qq(ant_pos[:,0],'m'),
                qq(ant_pos[:,1],'m'),
                qq(ant_pos[:,2],'m'))
    
    uvw = np.zeros((n_time,n_baseline,3))
    
    for i,t in enumerate(time_str):
        ra_dec = phase_center_ra_dec[i//f_pc_time,:]
        me = measures()
        #me.set_data_path(casa_data_dir)
        me.do_frame(tel_site)
        me.do_frame(me.epoch('utc', str(t)))
        me.do_frame(me.direction('J2000',qq(ra_dec[0], 'rad'),qq(ra_dec[1], 'rad')))
        
        # Converts from ITRF to "J2000":
        antpos_c_casa = me.as_baseline(antpos_casa)
        # Rotate into UVW frame
        
        
        antpos_uvw_casa = me.to_uvw(antpos_c_casa)['measure']
        
        #print(antpos_uvw_casa)
        ant_uvw = casa_to_astropy(antpos_uvw_casa)
        ant_uvw = ant_uvw.xyz.value.T

        uvw[i,:,:] = np.ascontiguousarray(ant_uvw[antenna1,:] - ant_uvw[antenna2,:])
    return uvw
'''
def calc_uvw_CALC(jpx_de421, ant_pos, mjd, phase_center_ra_dec,time_obj,delta = 0.00001):
    """
    Parameters
    ----------
    ant_pos : numpy.array (n_antx3), Geocentric ITRF, m
    mjd : numpy.array (n_time), Modified Julian Day, UTC
    phase_center_ra_dec : string
        Define the UVW frame relative to a certain point on the sky.
    Returns
    -------
    ant_uvw
    """
    
    #from calc11 import almacalc
    ref_antenna  = 0 #Choosing the first antenna as the reference
    n_times = len(mjd)

    ########################################################################################################
    #Geocentric (ITRF) position of each antenna.
    #ant_pos #n_ant x 3
    ant_x = np.ascontiguousarray(ant_pos[:,0])
    ant_y = np.ascontiguousarray(ant_pos[:,1])
    ant_z = np.ascontiguousarray(ant_pos[:,2])
    n_ant = ant_x.shape[0]
    #Geocentric position of the array reference point (ITRF).
    ref_x = ant_x[ref_antenna]
    ref_y = ant_y[ref_antenna]
    ref_z = ant_z[ref_antenna]
    ########################################################################################################
    # Temperature (deg. C), Pressure (hPa/mbar), and humidity (0-1) at each antenna
    #      REAL*8 temp(nant), pressure(nant), humidity(nant)
    # Only effects dry and wet delay
    temp = np.array([-1.68070068]*n_ant)  #To deg C, n_ant
    pressure = np.array([555.25872803]*n_ant) #Pressure (hPa/mbar), n_ant
    humidity = np.array([0.054894]*n_ant) #humidity (0-1), n_ant

    ########################################################################################################
    #phase_center_ra_dec radians, n_time x 2
    ra = np.ascontiguousarray(phase_center_ra_dec[:,0])
    dec = np.ascontiguousarray(phase_center_ra_dec[:,1])

    ssobj = np.zeros(n_times, dtype=bool) #True if the source is a solar system object.
    #Earth orientation parameters at each time (arc-sec, arc-sec, sec)
    iers_b = iers.IERS_B.open()
    dx_dy = np.ascontiguousarray(np.array(iers_b.pm_xy(time_obj))) #2 x n_time
    dx = np.ascontiguousarray(np.array(dx_dy[0,:]))
    dy = np.ascontiguousarray(np.array(dx_dy[1,:]))
    dut  = np.ascontiguousarray(np.array(iers_b.ut1_utc(time_obj)))
    print(ant_x)
    #print(n_times)

    leapsec = 35
    axisoff = np.zeros(n_ant)
    
    sourcename = np.array(['P'] * n_times) # source names, for future use with solar system objects
    #jpx_de421 = '~/sirius/sirius/DE421_little_Endian' #Path name of the JPL ephemeris
    
    #Calculate uvw using same math as DiFX (The software is availble at https://www.atnf.csiro.au/vlbi/dokuwiki/doku.php/difx/installation see Applications/calcif2/src/difxcalc.c::callCalc for math)
    geodelay, drydelay, wetdelay = almacalc(ref_x, ref_y, ref_z, ant_x, ant_y,ant_z, temp, pressure, humidity, mjd, ra, dec, ssobj,dx, dy, dut, leapsec, axisoff,sourcename, jpx_de421)
    ra_x = ra - delta/np.cos(dec)
    geodelay_x, drydelay_x, wetdelay_x = almacalc(ref_x, ref_y, ref_z, ant_x, ant_y,ant_z, temp, pressure, humidity, mjd, ra_x, dec, ssobj,dx, dy, dut, leapsec, axisoff,sourcename, jpx_de421)
    dec_y = dec + delta
    geodelay_y, drydelay_y, wetdelay_y = almacalc(ref_x, ref_y, ref_z, ant_x, ant_y,ant_z, temp, pressure, humidity, mjd, ra, dec_y, ssobj,dx, dy, dut, leapsec, axisoff,sourcename, jpx_de421)

    c = 299792458
    
    u = (c/delta)*(geodelay-geodelay_x)[:,0]
    v = (c/delta)*(geodelay_y-geodelay)[:,0]
    w = (c*geodelay)[:,0]
    

#    total_delay = geodelay + drydelay + wetdelay
#    total_delay_x = geodelay_x + drydelay_x + wetdelay_x
#    total_delay_y = geodelay_y + drydelay_y + wetdelay_y
#
#    u = (c/delta)*(total_delay-total_delay_x)[:,0]
#    v = (c/delta)*(total_delay_y-total_delay)[:,0]
#    w = (c*total_delay)[:,0]

    
    uvw = np.array([u,v,w]).T
    return uvw
'''



 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#Parallel writes https://github.com/ornladios/ADIOS2
#https://github.com/casacore/casacore/issues/432
#https://github.com/casacore/casacore/issues/729

import os
import dask.array as da
import dask
import xarray as xr
import numpy as np
from astropy import units as u

def simulation(point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, phase_center_names, beam_parms,beam_models,beam_model_map,uvw_parms, tel_xds, time_xda, chan_xda, pol, noise_parms, save_parms):
    """
    Simulate a interferometric visibilities and uvw coordinates. Dask enabled function
    
    Parameters
    ----------
    point_source_flux : np.array
    Returns
    -------
    vis : np.array
    uvw : np.array
    """

    from sirius import calc_vis, calc_uvw, calc_a_noise
    from sirius.make_ant_sky_jones import evaluate_beam_models
    from ._sirius_utils._array_utils import _ndim_list
    from ._parm_utils._check_beam_parms import _check_beam_parms
    from ._parm_utils._check_uvw_parms import _check_uvw_parms
    from ._parm_utils._check_save_parms import _check_save_parms
    from ._parm_utils._check_noise_parms import _check_noise_parms
    from ._sirius_utils._array_utils import _is_subset
    from ._sirius_utils._constants import pol_codes_RL, pol_codes_XY
    import numpy as np
    from itertools import cycle
    import itertools
    import copy
    
    
    _beam_parms = copy.deepcopy(beam_parms)
    _uvw_parms = copy.deepcopy(uvw_parms)
    _save_parms = copy.deepcopy(save_parms)
    _noise_parms = copy.deepcopy(noise_parms)
    assert(_check_uvw_parms(_uvw_parms)), "######### ERROR: calc_uvw uvw_parms checking failed."
    assert(_check_beam_parms(_beam_parms)), "######### ERROR: make_ant_sky_jones beam_parms checking failed."
    if noise_parms is not None:
        _noise_parms['freq_resolution'] = chan_xda.freq_resolution
        _noise_parms['time_delta'] = time_xda.time_delta
        _noise_parms['auto_corr'] = _uvw_parms['auto_corr']
        assert(_check_noise_parms(_noise_parms)), "######### ERROR: make_ant_sky_jones beam_parms checking failed."
    assert(_check_save_parms(_save_parms)), "######### ERROR: save_parms checking failed."
    
    pol = np.array(pol)
    assert(_is_subset(pol_codes_RL,pol) or _is_subset(pol_codes_XY,pol)), print('Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is ', pol)
    
    ### TO DO ###
    #Add checks dims n_time, n_chan, n_ant, n_point_sources are consistant or singleton when allowed.
    
    n_time = len(time_xda)
    n_chan = len(chan_xda)
    n_ant = tel_xds.dims['ant_name']
    #print(n_time,n_chan,n_ant)
    
    
    #Check all dims are either 1 or n
    f_pc_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
    f_ps_time = n_time if point_source_ra_dec.shape[0] == 1 else 1
    f_sf_time = n_time if point_source_flux.shape[1] == 1 else 1
    f_sf_chan = n_chan if point_source_flux.shape[2] == 1 else 1
    
    do_pointing = False
    if pointing_ra_dec is not None:
        do_pointing = True
        f_pt_time = n_time if phase_center_ra_dec.shape[0] == 1 else 1
        f_pt_ant =  n_ant if point_source_ra_dec.shape[1] == 1 else 1
    else:
        pointing_ra_dec = np.zeros((2,2,2))
        f_pt_time = n_time
        f_pt_ant = n_ant
    
    
    n_time_chunks = time_xda.data.numblocks[0]
    n_chan_chunks = chan_xda.data.numblocks[0]
    
    #Iter over time,chan
    iter_chunks_indx = itertools.product(np.arange(n_time_chunks), np.arange(n_chan_chunks))
    n_pol = len(pol)
    
    vis_list = _ndim_list((n_time_chunks,1,n_chan_chunks,1))
    uvw_list = _ndim_list((n_time_chunks,1,1))
    weight_list = _ndim_list((n_time_chunks,1,1))
    sigma_list = _ndim_list((n_time_chunks,1,1))
    
    from ._sirius_utils._array_utils import _calc_n_baseline
    n_baselines = _calc_n_baseline(n_ant,_uvw_parms['auto_corr'])
    
    # Build graph
    for c_time, c_chan in iter_chunks_indx:
        
        time_chunk = time_xda.data.partitions[c_time]
        chan_chunk = chan_xda.data.partitions[c_chan]
        
        #print(time_da.chunks[0][0])
        s_time = c_time*time_xda.data.chunks[0][0]
        e_time = c_time*time_xda.data.chunks[0][0] + time_xda.data.chunks[0][c_time] - 1 #-1 needed for // to work.
        s_chan = c_chan*chan_xda.data.chunks[0][0]
        e_chan = c_chan*chan_xda.data.chunks[0][0] + chan_xda.data.chunks[0][c_chan] - 1 #-1 needed for // to work.
        
        #print(s_time_indx,e_time_indx + 1)
        #print(s_time_indx//f_sf_time,e_time_indx//f_sf_time + 1)
        
        #point_source_flux: np.array [n_point_sources,n_time, n_chan, n_pol] (singleton: n_time, n_chan, n_pol)
        point_source_flux_chunk = point_source_flux[:,s_time//f_sf_time:e_time//f_sf_time+1,s_chan//f_sf_chan:e_chan//f_sf_chan+1,:]
        point_source_ra_dec_chunk = point_source_ra_dec[s_time//f_ps_time:e_time//f_ps_time+1,:,:]
        phase_center_ra_dec_chunk = phase_center_ra_dec[s_time//f_pc_time:e_time//f_pc_time+1,:]
        
        if do_pointing:
            pointing_ra_dec_chunk = pointing_ra_dec[s_time//f_pt_time:e_time//f_pt_time+1,:,:]
        else:
            pointing_ra_dec_chunk = None

        ### TO DO ###
        # Subselect channels for each beam_model with channel axis
        
        #print('time_chunk',time_chunk, chan_chunk)
    
        
        #print(c_time, c_chan)
        sim_chunk = dask.delayed(simulation_chunk)(
            dask.delayed(point_source_flux_chunk),
            dask.delayed(point_source_ra_dec_chunk),
            dask.delayed(pointing_ra_dec_chunk),
            dask.delayed(phase_center_ra_dec_chunk),
            dask.delayed(_beam_parms),beam_models,
            dask.delayed(beam_model_map),
            dask.delayed(_uvw_parms),
            tel_xds,
            time_chunk,
            chan_chunk,
            dask.delayed(pol), dask.delayed(_noise_parms),
            dask.delayed(None))
        #sim_chunk.compute()
        
        vis_list[c_time][0][c_chan][0] = da.from_delayed(sim_chunk[0],(len(time_chunk), n_baselines, len(chan_chunk),n_pol),dtype=np.complex)
        uvw_list[c_time][0][0] = da.from_delayed(sim_chunk[1],(len(time_chunk), n_baselines, 3),dtype=np.complex)
        weight_list[c_time][0][0] = da.from_delayed(sim_chunk[2],(len(time_chunk), n_baselines, n_pol),dtype=np.float)
        sigma_list[c_time][0][0] = da.from_delayed(sim_chunk[3],(len(time_chunk), n_baselines, n_pol),dtype=np.float)
        
    vis = da.block(vis_list)
    uvw = da.block(uvw_list)
    weight = da.block(weight_list)
    sigma = da.block(sigma_list)
    
    if _save_parms['DAG_name_vis_uvw_gen']:
        dask.visualize([vis,uvw],filename=_save_parms['DAG_name_vis_uvw_gen'])
        
    #Create simple xds with simulated vis, uvw, weight and sigma
    vis_xds = xr.Dataset()
    coords = {'time':time_xda.data,'chan': chan_xda.data, 'pol': pol}
    vis_xds = vis_xds.assign_coords(coords)
    
    vis_xds['DATA'] = xr.DataArray(vis, dims=['time','baseline','chan','pol'])
    vis_xds['UVW'] = xr.DataArray(uvw, dims=['time','baseline','uvw'])
    vis_xds['WEIGHT'] = xr.DataArray(weight, dims=['time','baseline','pol'])
    vis_xds['SIGMA'] = xr.DataArray(sigma, dims=['time','baseline','pol'])
    ###################
    
    write_to_ms(vis_xds, time_xda, chan_xda, pol, tel_xds, phase_center_names, phase_center_ra_dec, _uvw_parms['auto_corr'],_save_parms)
    
    return vis_xds
    

def simulation_chunk(point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, beam_parms,beam_models,beam_model_map,uvw_parms, tel_xds, time_chunk, chan_chunk, pol, _noise_parms, uvw_precompute=None):
    """
    Simulate a interferometric visibilities and uvw coordinates.
    
    Parameters
    ----------
    point_source_flux : np.array
    Returns
    -------
    vis : np.array
    uvw : np.array
    """

    from sirius import calc_vis, calc_uvw, calc_a_noise
    from sirius.make_ant_sky_jones import evaluate_beam_models
    import numpy as np
    
    #Calculate uvw coordinates
    if uvw_precompute is None:
        uvw, antenna1,antenna2 = calc_uvw(tel_xds, time_chunk, phase_center_ra_dec, uvw_parms,check_parms=False)
    else:
        from ._sirius_utils._array_utils import _calc_baseline_indx_pair
        n_ant = len(ant_pos)
        antenna1,antenna2=_calc_baseline_indx_pair(n_ant,uvw_parms['auto_corr'])
        uvw = uvw_precompute
          
    #Evaluate zpc files
    eval_beam_models, pa = evaluate_beam_models(beam_models,beam_parms,chan_chunk,phase_center_ra_dec,time_chunk,tel_xds.attrs['telescope_name'])
    
    #print(eval_beam_models)
#
    #Calculate visibilities
    #shape, point_source_flux, point_source_ra_dec, pointing_ra_dec, phase_center_ra_dec, antenna1, antenna2, n_ant, chan_chunk, pb_parms = calc_vis_tuple
    
    vis_data_shape =  np.concatenate((uvw.shape[0:2],[len(chan_chunk)],[len(pol)]))
    
    #print('pol',pol)
    vis =calc_vis(uvw,vis_data_shape,point_source_flux,point_source_ra_dec,pointing_ra_dec,phase_center_ra_dec,antenna1,antenna2,chan_chunk,beam_model_map,eval_beam_models, pa, pol, beam_parms['mueller_selection'])

    if _noise_parms is not None:
        vis, weight, sigma = calc_a_noise(vis,uvw,beam_model_map,eval_beam_models, antenna1, antenna2,_noise_parms)
    else:
        n_time, n_baseline, n_chan, n_pol = vis.shape
        weight = np.ones((n_time,n_baseline,n_pol))
        sigma = np.ones((n_time,n_baseline,n_pol))
    
    return vis, uvw, weight, sigma

    
def make_time_xda(time_start='2019-10-03T19:00:00.000',time_delta=3600,n_samples=10,n_chunks=4):
    """
    Create a time series xarray array.
    Parameters
    ----------
    -------
    time_da : dask.array
    """
    from astropy.timeseries import TimeSeries

    ts = np.array(TimeSeries(time_start=time_start,time_delta=time_delta*u.s,n_samples=n_samples).time.value)
    chunksize = int(np.ceil(n_samples/n_chunks))
    time_da = da.from_array(ts, chunks=chunksize)
    print('Number of chunks ', len(time_da.chunks[0]))
    
    time_xda = xr.DataArray(data=time_da,dims=["time"],attrs={'time_delta':float(time_delta)})
    
    return time_xda

def make_chan_xda(spw_name='sband',freq_start = 3*10**9, freq_delta = 0.4*10**9, freq_resolution=0.01*10**9, n_channels=3, n_chunks=3):
    """
    Create a time series xarray array.
    Parameters
    ----------
    -------
    time_da : dask.array
    """
    freq_chan = (np.arange(0,n_channels)*freq_delta + freq_start).astype(float) #astype(float) needed for interfacing with CASA simulator.
    chunksize = int(np.ceil(n_channels/n_chunks))
    chan_da = da.from_array(freq_chan, chunks=chunksize)
    print('Number of chunks ', len(chan_da.chunks[0]))
    
    chan_xda = xr.DataArray(data=chan_da,dims=["chan"],attrs={'freq_resolution':float(freq_resolution),'spw_name':spw_name, 'freq_delta':float(freq_delta)})
    return chan_xda
      
def write_to_ms(vis_xds, time_xda, chan_xda, pol, tel_xds, phase_center_names, phase_center_ra_dec, auto_corr,save_parms):
    from casatools import simulator
    from casatasks import mstransform
    sm = simulator()
    n_time, n_baseline, n_chan, n_pol = vis_xds.DATA.shape
    
    ant_pos = tel_xds.ANT_POS.values
    os.system('rm -rf ' + save_parms['ms_name'])
    sm.open(ms=save_parms['ms_name']);
    
    ###########################################################################################################################
    ## Set the antenna configuration
    sm.setconfig(telescopename=tel_xds.telescope_name,
                    x=ant_pos[:,0],
                    y=ant_pos[:,1],
                    z=ant_pos[:,2],
                    dishdiameter=tel_xds.DISH_DIAMETER.values,
                    mount=['alt-az'],
                    antname=list(tel_xds.ant_name.values),  #CASA can't handle an array of antenna names.
                    coordsystem='global',
                    referencelocation=tel_xds.site_pos[0]);
                    
    ## Set the polarization mode (this goes to the FEED subtable)
    from sirius._sirius_utils._constants import pol_codes_RL, pol_codes_XY, pol_str
    from sirius._sirius_utils._array_utils import _is_subset
    if _is_subset(pol_codes_RL,pol): #['RR','RL','LR','LL']
        sm.setfeed(mode='perfect R L', pol=['']);
    elif _is_subset(pol_codes_XY,pol): #['XX','XY','YX','YY']
        sm.setfeed(mode='perfect X Y', pol=['']);
    else:
        assert False, print('Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is ', pol)
    
    sm.setspwindow(spwname=chan_xda.spw_name,
               freq=chan_xda.data[0].compute(),
               deltafreq=chan_xda.freq_delta,
               freqresolution=chan_xda.freq_resolution,
               nchannels=len(chan_xda),
               refcode='LSRK',
               stokes=' '.join(pol_str[pol]));


    if auto_corr:
        sm.setauto(autocorrwt=1.0)
    else:
        sm.setauto(autocorrwt=0.0)
        
    import astropy
    from astropy.time import Time
    from astropy import units as u
    mjd = Time(time_xda.data[0:2].compute(), scale='utc')
    integration_time = (mjd[1]-mjd[0]).to('second')
    
    start_time = (mjd[0] - (integration_time/2 + 37*u.second)).mjd
    start_time_dict = {'m0': {'unit': 'd', 'value': start_time}, 'refer': 'UTC', 'type': 'epoch'}

    sm.settimes(integrationtime=integration_time.value,
             usehourangle=False,
             referencetime=start_time_dict);
        
    fields_set = []
    from collections import Counter
    field_time_count = Counter(phase_center_names)
    
    #print(field_time_count,phase_center_names)
    if len(phase_center_names) == 1: #Single field case
        field_time_count[list(field_time_count.keys())[0]] = n_time
    
    start_time = 0
    for i,ra_dec in enumerate(phase_center_ra_dec): #In future make phase_center_ra_dec a unique list
        if phase_center_names[i] not in fields_set:
            dir_dict = {'m0': {'unit': 'rad', 'value': ra_dec[0]}, 'm1': {'unit': 'rad', 'value': ra_dec[1]}, 'refer': 'J2000', 'type': 'direction'}
            sm.setfield(sourcename=phase_center_names[i],sourcedirection=dir_dict)
            fields_set.append(phase_center_names[i])
            
            stop_time = start_time + integration_time.value*field_time_count[phase_center_names[i]]
            sm.observe(sourcename=phase_center_names[i],
                spwname=chan_xda.spw_name,
                starttime= str(start_time) + 's',
                stoptime= str(stop_time) + 's')
            start_time = stop_time
         
    
    n_row = n_time*n_baseline
    
    #print(vis_data.shape)
    #print(n_row,n_time, n_baseline, n_chan, n_pol)
    
    #This code will most probably be moved into simulation if we get rid of row time baseline split.
    vis_data_reshaped = vis_xds.DATA.data.reshape((n_row, n_chan, n_pol))
    uvw_reshaped = vis_xds.UVW.data.reshape((n_row, 3))
    weight_reshaped = vis_xds.WEIGHT.data.reshape((n_row,n_pol))
    sigma_reshaped = vis_xds.SIGMA.data.reshape((n_row,n_pol))
    
    
    #weight_spectrum_reshaped = np.tile(weight_reshaped[:,None,:],(1,n_chan,1))
    
    
    
#    print(weight_reshaped.compute().shape)
#    print(sigma_reshaped.compute().shape)
#    print(weight_reshaped)
#    print(sigma_reshaped)
    
    #dask_ddid = da.full(n_row, 0, chunks=chunks['row'], dtype=np.int32)
    
    #print('vis_data_reshaped',vis_data_reshaped)
    
    from daskms import xds_to_table, xds_from_ms, Dataset
    
    #print('vis_data_reshaped.chunks',vis_data_reshaped.chunks)
    row_id = da.arange(n_row,chunks=vis_data_reshaped.chunks[0],dtype='int32')
    
    
    dataset = Dataset({'DATA': (("row", "chan", "corr"), vis_data_reshaped), 'CORRECTED_DATA': (("row", "chan", "corr"), vis_data_reshaped),'UVW': (("row","uvw"), uvw_reshaped), 'SIGMA': (("row","pol"), sigma_reshaped), 'WEIGHT': (("row","pol"), weight_reshaped),  'ROWID': (("row",),row_id)})
    #,'WEIGHT_SPECTRUM': (("row","chan","pol"), weight_spectrum_reshaped)
    ms_writes = xds_to_table(dataset, save_parms['ms_name'], columns="ALL")
    
    if save_parms['DAG_name_write']:
        dask.visualize(ms_writes,filename=save_parms['DAG_name_write'])
        
    if save_parms['write_to_ms']:
        dask.compute(ms_writes)
    

    sm.close()
    
    from casatasks import flagdata
    flagdata(vis=save_parms['ms_name'],mode='unflag')




#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import jit
import numba
import numpy as np


def _is_subset(arr1, arr2):
    '''
    Is arr2 a subset of arr1.
    '''
    for a in arr2:
        if a not in arr1:
            return False
    return True

def _ndim_list(shape):
    return [_ndim_list(shape[1:]) if len(shape) > 1 else None for _ in range(shape[0])]

#@jit(nopython=True,cache=True,nogil=True)
def _calc_baseline_indx_pair(n_ant,auto_corr):
    if auto_corr:
        n_baseline = int((n_ant**2 + n_ant)/2)
        a = 0
    else:
        n_baseline = int((n_ant**2 - n_ant)/2)
        a = 1
        
    antenna1 = np.zeros((n_baseline,),dtype=np.int)
    antenna2 = np.zeros((n_baseline,),dtype=np.int)
    
    k = 0
    for i in range(n_ant-a):
        for j in range(i+a,n_ant):
            antenna1[k] = i
            antenna2[k] = j
            k=k+1
            
    return antenna1, antenna2
    
def _calc_n_baseline(n_ant,auto_corr):
    if auto_corr:
        return int((n_ant**2 + n_ant)/2)
    else:
        return int((n_ant**2 - n_ant)/2)


#@jit("void(i8,i8,i8,i8,i8[:])",nopython=True,cache=True,nogil=True)
@jit(nopython=True,cache=True,nogil=True)
def _combine_indx_permutation(i1,i2,n_i1,n_i2):
    n_comb = n_i1*n_i2
    if n_i1 <= n_i2:
        i_comb = i1*n_i2 + i2
    else:
        i_comb = i1 + i2*n_i1
    
    #result[0] = i_comb
    #result[1] = n_comb
    #print(i1,i2,n_i1,n_i2,i_comb,n_comb)
    return i_comb,n_comb
    
#@jit("void(i8,i8,i8,i8,i8[:])",nopython=True,cache=True,nogil=True)
@jit(nopython=True,cache=True,nogil=True)
def _combine_indx_combination(i1,i2,n_i1,n_i2):
    if n_i1 <= n_i2:
        if i1 > i2:
            temp = i2
            i2 = i1
            i1 = temp
    
        n_comb = n_i1*n_i2 - (n_i1-1)*n_i1//2
        i_comb = ((2*n_i2 -1)*i1 - i1**2)//2 + i2
    else:
        if i1 < i2:
            temp = i2
            i2 = i1
            i1 = temp
    
        n_comb = n_i1*n_i2 - (n_i2-1)*n_i2//2
        i_comb = i1 + ((2*n_i1 -1)*i2 - i2**2)//2
        
    #print(i1,i2,n_i1,n_i2,i_comb,n_comb)
    return i_comb,n_comb

#The next set of functions finds the index of the value in a lists (val_list) that lies closest to a given val. These functions are needed for numba jit code.
@jit(nopython=True,cache=True,nogil=True)
def _find_val_indx(val_list,val):
    min_dif = -42.0 #Dummy value
    for jj in range(len(val_list)):
        ang_dif = np.abs(val-val_list[jj])

        if (min_dif < 0) or (min_dif > ang_dif):
            min_dif = ang_dif
            val_list_indx = jj
            
    return val_list_indx

#ang_list and ang must have the same anfle def. For example [-pi,pi] or [0,2pi]
@jit(nopython=True,cache=True,nogil=True)
def _find_angle_indx(ang_list,ang):
    min_dif = 42.0 #Dummy value
    for jj in range(len(ang_list)):
        #https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
        ang_dif = ang-ang_list[jj]
        ang_dif = np.abs((ang_dif + np.pi)%(2*np.pi) - np.pi)
        
        if min_dif > ang_dif:
            min_dif = ang_dif
            ang_list_indx = jj
            
    return ang_list_indx
    
#Uses an approximation for the distance between two pointings
@jit(nopython=True,cache=True,nogil=True)
def _find_ra_dec_indx(point_list,point):
    min_dis = 42.0 #Dummy value
    for jj in range(len(point_list)):
        #https://stjerneskinn.com/angular-distance-between-stars.htm
        #http://spiff.rit.edu/classes/phys373/lectures/radec/radec.html
        ra = point_list[jj,0]
        dec = point_list[jj,1]
        dis = np.sqrt(((ra-point[0])*np.cos(dec))**2 + (dec-point[1])**2) #approximation
        
        if min_dis > dis:
            min_dis = dis
            point_list_indx = jj

    return point_list_indx





#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#ducting - code is complex and might fail after some time if parameters is wrong. Sensable values are also checked. Gives printout of all wrong parameters. Dirty images alone has 14 parameters.

import numpy as np
from  ._check_parms import _check_parms, _check_dataset

def _check_uvw_parms(uvw_parms):
    import numbers
    parms_passed = True
    
    if not(_check_parms(uvw_parms, 'calc_method', [str], default = 'astropy',acceptable_data=['astropy','casa','casa_thread_unsafe'])): parms_passed = False
    if not(_check_parms(uvw_parms, 'auto_corr', [bool],default=False)): parms_passed = False
    
    return parms_passed



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import jit
import numba
import numpy as np

import os
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS

# Import required tools/tasks
from casatools import simulator, image, table, coordsys, measures, componentlist, quanta, ctsys, ms
from casatasks import tclean, imstat, visstat
from casatasks.private import simutil
from IPython.display import Markdown as md

# Instantiate all the required tools
sm = simulator()
ia = image()
tb = table()
cs = coordsys()
me = measures()
qa = quanta()
cl = componentlist()
mysu = simutil.simutil()
myms = ms()

import pylab as pl
import cngi.dio as dio
from cngi.conversion import convert_ms

def _display_image(imname='sim.image', pbname='', resname='',source_peak=1.0,chan=0,ylim=[0.4,1.1]):
    ia.open(imname)
    shp = ia.shape()
    csys = ia.coordsys()
    impix = ia.getchunk()
    ia.close()
    if pbname != '':
        ia.open(pbname)
        impb = ia.getchunk()
        ia.close()

    rad_to_deg =  180/np.pi
    w = WCS(naxis=2)
    w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
    w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
    w.wcs.crval = csys.referencevalue()['numeric'][0:2]*rad_to_deg
    w.wcs.ctype = ['RA---SIN','DEC--SIN']
    #w.wcs.ctype = ['RA','DEC']

    #pl.figure(figsize=(12,5))
    pl.figure(figsize=(12,5))
    pl.clf()
    #pl.subplot(121)
    pl.subplot(121,projection=w)

    p1 = shp[0]#int(shp[0]*0.25)
    p2 = shp[1]#int(shp[0]*0.75)

    if pbname != '':
        impix[impb < 0.1] = 0.0
    
    pl.imshow(impix[:,:,0,chan].transpose(), origin='lower')
    if pbname != '':
        pl.contour(impb[:,:,0,chan].transpose(),[0.2],colors=['magenta'], origin='lower')
    pl.title('Image from channel 0')
    pl.xlabel('Right Ascension')
    pl.ylabel('Declination')
    
    
    pk = 0.0
    if shp[3]>1:
        pl.subplot(122)
        ploc = np.where( impix == impix.max() )
        pl.plot(impix[ploc[0][0], ploc[1][0],0,:]/source_peak,'b*-',label='Im', markersize=18)
        if pbname != '':
            pl.plot(impb[ploc[0][0], ploc[1][0],0,:],'ro-',label='PB')
        pl.title('Spectrum at source peak')
        pl.xlabel('Channel')
        #pl.ylim((0.4,1.1))
        pl.legend()
        pk = impix[ploc[0][0], ploc[1][0],0,0]
        print('Peak Intensity (chan0) : %3.7f'%(pk))
        if pbname != '':
            pbk = impb[ploc[0][0], ploc[1][0],0,0]
            print('PB at location of Intensity peak (chan0) : %3.7f'%(pbk))
        pl.ylim(ylim)

    else:
        ploc = np.where( impix == impix.max() )
        print("Image Peak : %3.4f"%(impix[ploc[0][0], ploc[1][0],0,0]))
        if pbname != '':
            print("PB Value : %3.4f"%(impb[ploc[0][0], ploc[1][0],0,0]))
        pk = impix[ploc[0][0], ploc[1][0],0,0]

    if resname !='':
        istat = imstat(resname)  ### Make this calc within the PB.
        rres = istat['rms'][0]
        print('Residual RMS : %3.7f'%(rres))
    else:
        rres = None
    
 
    return pk, rres   # Return peak intensity from channnel 0 and rms


def _x_plot(vis='sim_data_ALMA.ms',ptype='amp-time',forceconvert=False,tel_name='ALMA'):
    """
    Make a few types of plots
    Supported types : amp-time, amp-freq, uvcov, plotants
    forceconvert=True/False : Convert the input MS to a Zarr dataset and read it into an XArray for plotting. 
                                               If set to False, it will skip the conversion step (and all the output messages the conversion produces). 
    """
    zvis = vis+'.zarr'
    if not os.path.exists(zvis) or forceconvert==True:
        convert_ms(vis, vis+'.zarr')
        
    xds = dio.read_vis(zvis)
    xdat = xds.xds0
    #print(xds)
    #print(xdat)
    #print(xds.ANTENNA)
    gxdat = xds.ANTENNA
    
    ant_names = gxdat.NAME.values
    xdat['field'] = xdat.FIELD_ID[:,0]
    
    
    xdat['DMAG'] = ((xdat['DATA'].real ** 2 + xdat['DATA'].imag ** 2) ** 0.5).mean(axis=3)
    xdat['U'] = xdat['UVW'][:,:,0]
    xdat['V'] = xdat['UVW'][:,:,1]
    xdat['-U'] = -xdat['UVW'][:,:,0]
    xdat['-V'] = -xdat['UVW'][:,:,1]
    
    ant_dias = np.unique(gxdat.DISH_DIAMETER)

    xAA = xdat.where( (gxdat.DISH_DIAMETER[xdat.ANTENNA1] == ant_dias[0])  &  (gxdat.DISH_DIAMETER[xdat.ANTENNA2] == ant_dias[0])  )
    xBB = xdat.where(  (gxdat.DISH_DIAMETER[xdat.ANTENNA1] == ant_dias[1])  &  (gxdat.DISH_DIAMETER[xdat.ANTENNA2] == ant_dias[1])  )
    xAB = xdat.where( ( (gxdat.DISH_DIAMETER[xdat.ANTENNA1] == ant_dias[0])  &  (gxdat.DISH_DIAMETER[xdat.ANTENNA2] == ant_dias[1]) | (gxdat.DISH_DIAMETER[xdat.ANTENNA1] == ant_dias[1])  &  (gxdat.DISH_DIAMETER[xdat.ANTENNA2] == ant_dias[0]) ) )


    if ptype == 'amp-time':
        fig, axes = pl.subplots(ncols=1,figsize=(9,3))
        for fld in np.unique(xdat.field):
            xAA.where(xAA.field==fld).plot.scatter(x='time',y='DMAG',  marker='.', color='r',alpha=0.1,label='A-A')
            xAB.where(xAB.field==fld).plot.scatter(x='time',y='DMAG',  marker='.', color='m',alpha=0.1,label='A-B')
            xBB.where(xBB.field==fld).plot.scatter(x='time',y='DMAG',  marker='.', color='b',alpha=0.1,label='B-B')
            
        pl.title('Visibility ampllitude : Red (A-A), Blue (B-B), Purple (A-B)');

    if ptype == 'amp-freq':
        fig, axes = pl.subplots(ncols=2,figsize=(9,3))
        ax=0
        for fld in np.unique(xdat.field):
            xAA.where(xAA.field==fld).plot.scatter(x='chan',y='DMAG',  marker='.', color='r',alpha=0.1,ax=axes[ax])
            xAB.where(xAB.field==fld).plot.scatter(x='chan',y='DMAG',  marker='.', color='m',alpha=0.1,ax=axes[ax])
            xBB.where(xBB.field==fld).plot.scatter(x='chan',y='DMAG',  marker='.', color='b',alpha=0.1,ax=axes[ax])
            axes[ax].set_title('Visibility Spectrum for field : '+ str(fld)) #+ '\nRed (A-A), Blue (B-B), Purple (A-B)')
            ax = ax+1
            
    if ptype == 'uvcov':
        fig, axes = pl.subplots(ncols=2,figsize=(9,4))
        ant_dias = np.unique(gxdat.DISH_DIAMETER)
        ax=0
        for fld in np.unique(xdat.field):
            xAB.where(xAB.field==fld).plot.scatter(x='U',y='V',  marker='.', color='m',ax=axes[ax])
            xAB.where(xAB.field==fld).plot.scatter(x='-U',y='-V',  marker='.', color='m',ax=axes[ax])
            xBB.where(xBB.field==fld).plot.scatter(x='U',y='V',  marker='.', color='b', ax=axes[ax])
            xBB.where(xBB.field==fld).plot.scatter(x='-U',y='-V',  marker='.', color='b', ax=axes[ax])
            xAA.where(xAA.field==fld).plot.scatter(x='U',y='V',  marker='.', color='r', ax=axes[ax])
            xAA.where(xAA.field==fld).plot.scatter(x='-U',y='-V',  marker='.', color='r', ax=axes[ax])
            axes[ax].set_title('UV coverage for field : '+str(fld))
            ax=ax+1
 

    if ptype == 'plotants':
        if tel_name=='ALMA':
            typeA = 'A'
        else:
            typeA = 'm'
        fig, axes = pl.subplots(ncols=1,figsize=(6,5))
        gxdat['ANT_XPOS'] = gxdat['POSITION'][:,0] - gxdat['POSITION'][:,0].mean()
        gxdat['ANT_YPOS'] = gxdat['POSITION'][:,1] - gxdat['POSITION'][:,1].mean()
        gxdat.plot.scatter(x='ANT_XPOS', y='ANT_YPOS',color='k',marker="1",s=200,linewidth=3.0)
        
        for i, txt in enumerate(ant_names):
            col = ('b' if (txt.count(typeA)>0) else 'r')
            pl.annotate('   '+txt, (gxdat['ANT_XPOS'].values[i], gxdat['ANT_YPOS'].values[i]),fontsize=12,color=col)   
        pl.title('Antenna Positions')
        
def _listobs_jupyter(vis='sim_data_ALMA.ms'):
    """
    Print out the contents of listobs.
    TODO : Convert the contents to a df (if possible) for a pretty display
    """
    from casatasks import listobs
    listobs(vis=vis, listfile='obslist.txt', verbose=False, overwrite=True)
    ## print(os.popen('obslist.txt').read()) # ?permission denied?
    fp = open('obslist.txt')
    for aline in fp.readlines():
        print(aline.replace('\n',''))
    fp.close()

    tb.open(vis+'/ANTENNA')
    print("Dish diameter : " + str(tb.getcol('DISH_DIAMETER')))
    print("Antenna name : " + str(tb.getcol('NAME')))
    tb.close()
        
        
        
def _image_ants(vis='sim_data_ALMA.ms',imname='try_ALMA', field='0', antsel='',tel='ALMA',vptable=''):
    """
    Run imager.....
    TODO  set pblimit=-1 ... or set the image stats method to look inside the PB region only.
    """
    antsels =  _get_baseline_types(tel=tel)

    if antsel not in antsels.keys():
        print('Pick an antenna selection from '+ str(antsels))
        return
   
    
    if tel=='ALMA':
        cell='0.3arcsec'
        imsize=1024
        phasecenter='J2000 +19h59m28.5s -40d44m21.5s'
    if (tel=='NGVLA'):
        cell='1.0arcsec'
        imsize=1024
        phasecenter='J2000 +19h59m28.5s +40d44m21.5s'

   
    if field=='0' or field=='1':
        ftype = 'single'
    else:
        ftype = 'mosaic'
    
    imname1 = imname+'_'+antsel+'_'+ftype
    
    os.system('rm -rf '+imname1+'.*')
    
    print('antsels[antsel]2 ',antsels[antsel])
    
    # Run tclean
    tclean(vis=vis,
       antenna=antsels[antsel],
       field=field,
       imagename=imname1,
       imsize=imsize,
       cell=cell,
       phasecenter=phasecenter,
       specmode='cube',
       interpolation='nearest',
       nchan=-1,
       gridder='mosaic',
       vptable=vptable,
       normtype='flatnoise',
       wbawp=True,
       #pblimit=0.05,
       pblimit=0.2,
       pbmask=0.2,
       conjbeams=False,
       niter=1000,
       nsigma=3.0,
       datacolumn='data',
       weighting='natural')
        
        
        
def _get_baseline_types(msname='',tel='ALMA'):
    """
    Define MSSelection strings for all 3 types of baselines, and all.
    Check them if an MS name is supplied.
    """

    if tel=='ALMA':
        antsels = {'A':'A*&',
                   'B':'J*,N*&',
                   'cross':'A* && J*,N*',
                   'all':'*'}
    if (tel=='NGVLA' or tel=='KSA'):
        antsels = {'A':'m*&',
                   'B':'s*&',
                   'cross':'m* && s*',
                   'all':'*'}


    if msname != "":
        for atype in antsels.keys():
            asel = myms.msseltoindex(vis=msname, baseline=antsels[atype])
            print(antsels[atype])
            print(asel['antenna1'])
            print(asel['antenna2'])
            print(asel['baselines'].transpose())
   
    return antsels
        


def _check_vals(vis='',field='',spw='',pb_A='', pb_B='', antsels={}, meas_peaks={}, meas_rms={},tel='ALMA'):
    """
    For the given (selected) MS and PB images, pick out the PB values at the source location for antenna types A and B
    Calculate expected Intensity for the A-A, B-B, A-B and All baseline imaging cases.
    Using sigmas (or weights) and input vis-rms values, calculate expected rms for all 4 imaging cases
    Compare expected intensity, PB and RMS values with measured values.
    
    ALMA : A = 12m.   B = 7m
    ngVLA : A = 18m.  B = 6m
   
    """

    if tel=='ALMA':
        D_A = 12.0
        D_B = 7.0
        label_A = '12m'
        label_B = '7m'
        freq = 90.0 # GHz
    if (tel=='NGVLA'):
        D_A = 18.0
        D_B = 6.0
        label_A = '18m'
        label_B = '6m'
        freq = 40.0 # GHz
    if (tel=='KSA'):
        D_A = 20.0
        D_B = 10.0
        label_A = '20m'
        label_B = '10m'
        freq = 40.0 # GHz

    antsels3 = _get_baseline_types(tel=tel)
    antsels3.pop('all')
    vis_summ = _get_vis_stat(vis=vis,field=field,spw=spw, antsels=antsels3)
    
    calc_peak_A = vis_summ['A']['mean']
    calc_peak_B = vis_summ['B']['mean']
    calc_peak_cross = vis_summ['cross']['mean']
    
    
    calc_peak_all = ( vis_summ['A']['mean']* vis_summ['A']['meanwt'] *  vis_summ['A']['npts'] +                \
                              vis_summ['B']['mean']* vis_summ['B']['meanwt']* vis_summ['B']['npts'] +                        \
                              vis_summ['cross']['mean']* vis_summ['cross']['meanwt']* vis_summ['cross']['npts'] ) /               \
                                 (vis_summ['A']['npts']* vis_summ['A']['meanwt'] +                                                       \
                                  vis_summ['B']['npts']* vis_summ['B']['meanwt'] +                                                           \
                                  vis_summ['cross']['npts']* vis_summ['cross']['meanwt'])
    
    calc_rms_A = (vis_summ['A']['stddev']/np.sqrt(vis_summ['A']['npts']) )
    calc_rms_B = vis_summ['B']['stddev']/np.sqrt(vis_summ['B']['npts'])
    calc_rms_cross = vis_summ['cross']['stddev']/np.sqrt(vis_summ['cross']['npts'])
    
    sumwt = (vis_summ['A']['npts']* vis_summ['A']['meanwt'] +                                                       \
                    vis_summ['B']['npts']* vis_summ['B']['meanwt'] +                                                           \
                    vis_summ['cross']['npts']* vis_summ['cross']['meanwt'])

    calc_rms_all = np.sqrt(( (calc_rms_A * vis_summ['A']['meanwt'] *  vis_summ['A']['npts']/sumwt)**2 +              \
                                         (calc_rms_B * vis_summ['B']['meanwt']* vis_summ['B']['npts']/sumwt)**2 +                      \
                                         (calc_rms_cross * vis_summ['cross']['meanwt']* vis_summ['cross']['npts']/sumwt)**2 )  )
  


    normn_A = vis_summ['A']['stddev']/vis_summ['A']['stddev']
    normn_B = vis_summ['B']['stddev']/vis_summ['A']['stddev']
    normn_cross = vis_summ['cross']['stddev']/vis_summ['A']['stddev']

    import pandas as pd
    from IPython.display import display, HTML
    #pd.set_option("display.precision", 3)
    results = { 'Baseline\nTypes' : [label_A+'-'+label_A,label_B+'-'+label_B , label_A+'-'+label_B, 'All'],
                'Vis Mean \n(V=I=P)' : [vis_summ['A']['mean'],vis_summ['B']['mean'], vis_summ['cross']['mean'], "-NA-" ],
           'Vis Std\n(sim)' : [vis_summ['A']['stddev'],vis_summ['B']['stddev'], vis_summ['cross']['stddev'], "-NA-" ],
                'Vis Std\n(norm sim)' : [normn_A, normn_B, normn_cross, "-NA-" ],
                'Vis Std\n(calc)' : [ 1.0, (D_A**2)/(D_B**2), (D_A**2)/(D_A*D_B)  , "-NA-" ],
           'Number of \n Data Pts' : [vis_summ['A']['npts'],vis_summ['B']['npts'],vis_summ['cross']['npts'],"-NA"],
           'Weight\n(calc)' :  [1.0, ((D_B**2)/(D_A**2))**2, ((D_A*D_B)/(D_A**2))**2 , "-NA-" ],
                'Weight\n(sim)' :[vis_summ['A']['meanwt'],vis_summ['B']['meanwt'], vis_summ['cross']['meanwt'], "-NA-" ],
           'Int Jy/bm\n(calc)' : [calc_peak_A, calc_peak_B, np.sqrt(calc_peak_A * calc_peak_B)  , calc_peak_all],
           'Int Jy/bm\n(meas)' : [meas_peaks['A'],meas_peaks['B'],meas_peaks['cross'],meas_peaks['all']],
           'RMS mJy\n(calc)' : [calc_rms_A*1e+3, calc_rms_B*1e+3, calc_rms_cross*1e+3, calc_rms_all*1e+3],
           'RMS mJy\n(meas)' : [meas_rms['A']*1e+3,meas_rms['B']*1e+3,meas_rms['cross']*1e+3,meas_rms['all']*1e+3]   }
    df = pd.DataFrame(results)
    with pd.option_context('display.float_format', '{:0.4f}'.format):
        display(HTML(df.to_html()))
        #print(pd.DataFrame(df).to_markdown())
    #print(df)
    

    print("Calculated PB size for type A (dia=%2.2f) : %3.5f arcmin"%(D_A, _calc_ang(freq,D_A)))
    print("Calculated PB size for type B (dia=%2.2f) : %3.5f arcmin"%(D_B, _calc_ang(freq,D_B)))
    

    return
    


def _get_vis_stat(vis='sim_data_ALMA.ms',field='',spw='', antsels={}):
    """
    ### This cell examines the dataset and checks noise level per baseline type.
    ## Use visstat with antenna selection ?
    """

    vis_summ = {}
    
    for btype in antsels.keys():
        vrec = visstat(vis=vis,field=field,spw=spw,antenna=antsels[btype],axis='amp',datacolumn='data')['DATA_DESC_ID=0']
        wrec = visstat(vis=vis,field=field,spw=spw,antenna=antsels[btype],axis='weight_spectrum',datacolumn='data')['DATA_DESC_ID=0']
        #print(vrec)
        vis_summ[btype] = {'mean':vrec['mean'], 'stddev':vrec['stddev'], 'sumwt':vrec['sumOfWeights'], 'npts':vrec['npts'], 'meanwt':wrec['mean']}
    
 #   df = pd.DataFrame(vis_summ)
 #   print("Summary of stats from the input MS")
 #   with pd.option_context('display.float_format', '{:0.8f}'.format):
 #       print(df)
 #   print("\n")
    return vis_summ

def _calc_ang(freq, dia):
    return  ( (3e+8/(freq*1e+9)) / dia ) * (180.0/np.pi) * 60.0



"""
SIRIUS functions
"""
#__init__.py
from .calc_a_noise import calc_a_noise
from .calc_vis import calc_vis
from .calc_uvw import calc_uvw
from .make_ant_sky_jones import make_ant_sky_jones
from .simulation import simulation




 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import numpy as np
c = 299792458
from ._sirius_utils._direction_rotate import _calc_rotation_mats, _cs_calc_rotation_mats
from ._sirius_utils._apply_primary_beam import _apply_casa_airy_pb
from ._sirius_utils._ant_jones_term import _calc_ant_jones
from ._sirius_utils._calc_parallactic_angles import _calc_parallactic_angles, _find_optimal_set_angle
import itertools
import matplotlib.pyplot as plt
from PIL import Image
import xarray as xr
import copy
from ._parm_utils._check_beam_parms import _check_beam_parms


def evaluate_beam_models(beam_models,beam_parms,freq_chan,phase_center_ra_dec,time_str,site):
    pa = _calc_parallactic_angles(time_str,site,phase_center_ra_dec)
    pa_subset,vals_dif = _find_optimal_set_angle(pa[:,None],beam_parms['pa_radius'] )
    
    _beam_parms = copy.deepcopy(beam_parms)
    _beam_parms['pa'] =  pa_subset
    _beam_parms['freq'] = freq_chan
    
    eval_beam_models = []
    for bm in beam_models:
        if 'ZC' in bm: #check for zpc files
            J_xds = make_ant_sky_jones([bm],_beam_parms) #[None,None,:,:,:,:]
            J_xds.attrs = bm.attrs
            eval_beam_models.append(J_xds)
        else:
            eval_beam_models.append(bm)
    
    return eval_beam_models, pa
    


#def create_pb(a_parm_indx,list_zpc_dataset,gcf_a_freq,gcf_a_pa,gcf_parms,grid_parms):
def make_ant_sky_jones(list_zpc_dataset,beam_parms):
    """
    Simulate a interferometric visibilities and uvw coordinates.
    
    Parameters
    ----------
    point_source_flux : np.array
    Returns
    -------
    vis : np.array
    uvw : np.array
    """

    _beam_parms = copy.deepcopy(beam_parms)
    
    pb_freq = _beam_parms['freq']
    pb_pa = _beam_parms['pa']
    
    min_delta = _calc_resolution(pb_freq,list_zpc_dataset,_beam_parms)
    #print('min_delta',min_delta)
    _beam_parms['cell_size'] = np.array([-min_delta,min_delta]) #- sign?
  
    map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
    _beam_parms['needed_pol'] = np.unique(np.ravel(map_mueler_to_pol[_beam_parms['mueller_selection']]))
    
    assert (0 in _beam_parms['mueller_selection']) or (15 in _beam_parms['mueller_selection']), "Mueller element 0 or 15 must be selected."
    
    pb_planes = _calc_ant_jones(list_zpc_dataset,pb_freq,pb_pa,_beam_parms)
    
    image_size = _beam_parms['image_size']
    image_center = image_size//2
    cell_size = _beam_parms['cell_size']
    
    image_center = np.array(image_size)//2
    l = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    m = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]
    
    coords = {'chan':pb_freq, 'pa': pb_pa, 'pol': _beam_parms['needed_pol'],'l':l,'m':m}

    J_xds = xr.Dataset()
    J_xds = J_xds.assign_coords(coords)
    
    J_xds['J'] = xr.DataArray(pb_planes, dims=['model','pa','chan','pol','l','m'])
    

    return J_xds
    
    
def _calc_resolution(pb_freq,list_zpc_xds,beam_parms):
    c = 299792458
    

    min_delta = 42
    for zpc_xds in list_zpc_xds:
        fov = beam_parms['fov_scaling']*(1.22 * c / (zpc_xds.dish_diam*pb_freq))
        delta = min(min(fov/beam_parms['image_size'][0]),min(fov/beam_parms['image_size'][1]))
        if delta < min_delta:
            min_delta =  delta
    return min_delta



#Sekhar et al 2020 in Preperation explains model.
#The column headings are assumed to be stokes,freq,ind,real,imag
def csv_to_zarr(filename,freq_to_hertz,dish_diam):
    #from cngi.dio import write_zarr
    import xarray as xr
    import numpy as np
    import dask.array as da
    import csv
    from datetime import date
    
    csv_list = []
    with open(filename, newline='') as csvfile:
        csvreader = csv.reader(csvfile, delimiter=',', quotechar='|')
        next(csvreader) #Heading ['#stokes,freq,ind,real,imag']
        for row in csvreader:
            csv_list.append(np.array(row,float))
            
    csv_array = np.array(csv_list)

    pol_codes = np.unique(csv_array[:,0]).astype(int)
    chan_freq = np.unique(csv_array[:,1])
    num_coef = int(np.max(csv_array[:,2]))+1
    
    zc_array =  np.zeros((len(pol_codes),len(chan_freq),num_coef),dtype=np.complex)
    eta_array = np.zeros((len(pol_codes),len(chan_freq),num_coef),dtype=np.double)
    
    for i_csv in range(len(csv_array)):
        i_pol = np.where(pol_codes == csv_array[i_csv,0])[0][0]
        i_chan = np.where(chan_freq == csv_array[i_csv,1])[0][0]
        i_coef = int(csv_array[i_csv,2])
        zc_array[i_pol,i_chan,i_coef] =  csv_array[i_csv,3] + 1j*csv_array[i_csv,4]
        #print(len(csv_array[i_csv,:]))
        if len(csv_array[i_csv,:]) > 5:
            eta_array[i_pol,i_chan,i_coef] = csv_array[i_csv,5]
        else:
            eta_array[i_pol,i_chan,i_coef] = 1
        
    zc_dataset = xr.Dataset()
    coords = {'pol':pol_codes,'chan':chan_freq*freq_to_hertz,'coef_indx':np.arange(num_coef)}
    zc_dataset = zc_dataset.assign_coords(coords)
    zc_dataset['ZC'] = xr.DataArray(zc_array, dims=['pol','chan','coef_indx'])
    zc_dataset['ETA'] = xr.DataArray(eta_array, dims=['pol','chan','coef_indx'])
    zc_dataset.attrs['name'] = filename
    zc_dataset.attrs['conversion_date'] = str(date.today())
    zc_dataset.attrs['dish_diam'] = dish_diam
    
    #write_zarr(zc_dataset,filename.split('.')[0]+'.zpc.zarr')
    xr.Dataset.to_zarr(zc_dataset,filename.split('.')[0]+'.zpc.zarr',mode='w')
    
    
    #print(zc_dataset)
    


if __name__ == '__main__':
    import shutil
    #Remove all . in name except for last (before .csv)
    #'data/EVLA_avg_SBand_coeffs_highoversamp_wideband.csv', EVLA_avg_zcoeffs_SBand_full_pol_lookup.csv
    filenames = ['EVLA_avg_zcoeffs_LBand_lookup.csv','EVLA_avg_zcoeffs_SBand_lookup.csv', 'MeerKAT_avg_zcoeffs_LBand_lookup.csv','EVLA_avg_zcoeffs_SBand_lookup_holoeta.csv','EVLA_avg_zcoeffs_SBand_lookup_noeta.csv','EVLA_avg_zcoeffs_SBand_lookup_stdeta.csv']
    dish_diams = [25,25,25,13.5,25]
    freq_to_hertz = 10**6
    for filename,dish_diam in zip(filenames,dish_diams):
        print(filename)
        csv_to_zarr('data/'+filename,freq_to_hertz,dish_diam)
        try:
            shutil.make_archive('data/'+filename[:-4]+'.zpc.zarr', 'zip', 'data/'+filename[:-4]+'.zpc.zarr')
        except:
            print('Cant compress','data/'+filename[:-4]+'.zpc.zarr')
